'use strict';

var Promise = require('bluebird');
var Errors = require('common-errors');
var scrypt = require('../utils/scrypt.js');
var redisKey = require('../utils/key.js');
var jwt = require('../utils/jwt.js');

module.exports = function login(opts) {
  var config = this.config.jwt;
  var redis = this.redis;
  var username = opts.username;
  var password = opts.password;
  var lockAfterAttempts = config.lockAfterAttempts;
  var defaultAudience = config.defaultAudience;

  var audience = opts.audience || defaultAudience;
  var remoteip = opts.remoteip || false;
  var usernameKey = redisKey(username, 'data');

  var promise = Promise.bind(this);
  var loginAttempts = undefined;
  if (remoteip && lockAfterAttempts > 0) {
    (function () {
      var remoteipKey = redisKey(usernameKey, remoteip);
      promise = promise.then(function checkLoginAttempts() {
        // construct pipeline
        var pipeline = redis.pipeline();

        pipeline.incrby(remoteipKey, 1);
        if (config.keepLoginAttempts > 0) {
          pipeline.expire(remoteipKey, config.keepLoginAttempts);
        }

        return pipeline.exec().spread(function incremented(incrementValue) {
          var err = incrementValue[0];
          if (err) {
            this.log.error('Redis error:', err);
            return;
          }

          loginAttempts = incrementValue[1];
          if (loginAttempts > lockAfterAttempts) {
            throw new Errors.HttpStatusError(429, 'You are locked from making login attempts for the next 24 hours');
          }
        });
      });
    })();
  } else {
    promise = Promise.resolve();
  }

  return promise.then(function getHashedPasswordAndUserState() {
    return redis.hmgetBuffer(usernameKey, 'password', 'active', 'banned');
  }).spread(function hashedPasswordAndUserState(passwordHashBuffer, activeBuffer, isBanned) {
    if (!passwordHashBuffer) {
      throw new Errors.HttpStatusError(404, 'username does not exist');
    }

    return scrypt.verify(passwordHashBuffer, password).then(function verificationResult(matches) {
      if (matches !== true) {
        throw new Errors.HttpStatusError(403, 'incorrect password');
      }

      if (remoteip) {
        loginAttempts = 0;
        return redis.del(remoteip);
      }
    }).then(function checkAccountActivation() {
      if (!activeBuffer || activeBuffer.toString() !== 'true') {
        throw new Errors.HttpStatusError(412, 'Account hasn\'t been activated');
      }

      if (isBanned && isBanned.toString() === 'true') {
        throw new Errors.HttpStatusError(423, 'Account has been locked');
      }
    });
  }).then(function getUserInfo() {
    return jwt.login.call(this, username, audience);
  })['catch'](function enrichError(err) {
    if (remoteip) {
      err.loginAttempts = loginAttempts;
    }

    throw err;
  });
};