'use strict';

var Promise = require('bluebird');
var Errors = require('common-errors');
var request = require('request-promise');
var ld = require('lodash');
var setMetadata = require('./updateMetadata.js');
var scrypt = require('../utils/scrypt.js');
var redisKey = require('../utils/key.js');
var emailValidation = require('../utils/send-email.js');
var jwt = require('../utils/jwt.js');

var _require = require('util');

var fmt = _require.format;

/**
 * Registration handler
 * @param  {Object} message
 * @return {Promise}
 */
module.exports = function registerUser(message) {
  var redis = this.redis;
  var config = this.config;
  var deleteInactiveAccounts = config.deleteInactiveAccounts;
  var captchaConfig = config.captcha;
  var secret = captchaConfig.secret;
  var ttl = captchaConfig.ttl;
  var uri = captchaConfig.uri;

  // message
  var username = message.username;
  var password = message.password;
  var audience = message.audience;
  var ipaddress = message.ipaddress;

  var activate = message.hasOwnProperty('activate') ? message.activate : true;
  var captcha = message.hasOwnProperty('captcha') ? message.captcha : false;
  var metadata = message.hasOwnProperty('metadata') ? message.metadata : false;

  // task holder
  var logger = this.log.child({ username: username, action: 'register' });

  var promise = Promise.bind(this);

  // optional captcha verification
  if (captcha) {
    logger.debug('verifying captcha');

    promise = promise.then(function checkCaptcha() {
      var captchaCacheKey = captcha.response;
      return redis.pipeline().set(captchaCacheKey, username, 'EX', ttl, 'NX').get(captchaCacheKey).exec().spread(function captchaCacheResponse(setResponse, getResponse) {
        if (getResponse[1] !== username) {
          throw new Errors.HttpStatusError(412, 'Captcha challenge you\'ve solved can not be used, please complete it again');
        }
      }).then(function verifyGoogleCaptcha() {
        return request.post({ uri: uri, qs: ld.defaults(captcha, { secret: secret }), json: true }).then(function captchaSuccess(body) {
          if (!body.success) {
            return Promise.reject({ statusCode: 200, error: body });
          }

          return true;
        })['catch'](function captchaError(err) {
          var errData = JSON.stringify(ld.pick(err, ['statusCode', 'error']));
          throw new Errors.HttpStatusError(412, fmt('Captcha response: %s', errData));
        });
      });
    });
  }

  if (ipaddress) {}
  // TODO:
  // add reg per ip address limits

  // shared user key
  var userDataKey = redisKey(username, 'data');

  // step 2, verify that user _still_ does not exist
  promise = promise.then(function userDoesntExist() {
    logger.debug('Verifying user existance');

    return redis.hexists(userDataKey, 'password').then(function userExists(exists) {
      if (exists) {
        throw new Errors.HttpStatusError(403, '"' + username + '" already exists');
      }

      return true;
    });
  });

  // step 3 - encrypt password
  promise = promise.tap(function logProgress() {
    logger.debug('hashing password');
  })['return'](password).then(scrypt.hash);

  // step 4 - create user if it wasn't created by some1 else trying to use race-conditions
  promise = promise.then(function createUser(hash) {
    logger.debug('inserting user');

    var pipeline = redis.pipeline();

    pipeline.hsetnx(userDataKey, 'password', hash);
    pipeline.hsetnx(userDataKey, 'active', activate);

    return pipeline.exec().spread(function insertedUserData(passwordSetResponse) {
      if (passwordSetResponse[1] === 0) {
        throw new Errors.HttpStatusError(403, 'User "' + username + '" already exists');
      }

      if (!activate && deleteInactiveAccounts >= 0) {
        // WARNING: IF USER IS NOT VERIFIED WITHIN <deleteInactiveAccounts>[by default 30] DAYS - IT WILL BE REMOVED FROM DATABASE
        return redis.expire(userDataKey, deleteInactiveAccounts);
      }
    });
  });

  // step 5 - save metadata if present
  promise = promise.then(function insertMetadata() {
    if (!metadata) {
      return null;
    }

    logger.debug('inserting metadata');
    return setMetadata.call(this, {
      username: username,
      audience: audience,
      metadata: {
        $set: metadata
      }
    });
  });

  // step 6 - generate JWT token or return success
  promise = promise.then(function generateTokenAndUserObject() {
    if (!activate) {
      logger.debug('sending activation email');

      emailValidation.send.call(this, username);

      return {
        requiresActivation: true
      };
    }

    logger.debug('returning activated user');
    return jwt.login.call(this, username, audience);
  });

  return promise;
};