'use strict';

var Errors = require('common-errors');
var scrypt = require('../utils/scrypt.js');
var redisKey = require('../utils/key.js');
var jwt = require('../utils/jwt.js');
var emailChallenge = require('../utils/send-email.js');

/**
 * Verifies token and deletes it if it matches
 * @param {Strong} token
 */
function tokenReset(token) {
  return emailChallenge.verify.call(this, token, 'reset', true);
}

/**
 * Verify that username and password match
 * @param {String} username
 * @param {String} password
 */
function usernamePasswordReset(username, password) {
  var redis = this.redis;

  var userKey = redisKey(username, 'data');
  return redis.hmgetBuffer(userKey, 'password', 'active').spread(function responses(hash, active) {
    if (!hash) {
      throw new Errors.HttpStatusError(404, 'user does not exist');
    }

    if (String(active) !== 'true') {
      throw new Errors.HttpStatusError(412, 'account is not active or does not exist');
    }

    if (!Buffer.isBuffer(hash) || hash.length < 1) {
      throw new Errors.HttpStatusError(500, 'invalid password hash');
    }

    return scrypt.verify(hash, password);
  })['return'](username);
}

module.exports = function updatePassword(opts) {
  var redis = this.redis;
  var password = opts.newPassword;
  var remoteip = opts.remoteip;

  var invalidateTokens = !!opts.invalidateTokens;

  // 2 cases - token reset and current password reset
  var promise = undefined;
  if (opts.resetToken) {
    promise = tokenReset.call(this, opts.resetToken);
  } else {
    promise = usernamePasswordReset.call(this, opts.username, opts.currentPassword);
  }

  promise = promise.then(function prepareData(username) {
    return scrypt.hash(password).then(function calculatedHash(hash) {
      var userKey = redisKey(username, 'data');
      return redis.hset(userKey, 'password', hash).then(function updatedPassword(result) {
        if (result !== 1) {
          return null;
        }

        return redis.hdel(userKey, 'password').done(function reportError() {
          throw new Errors.HttpStatusError(404, 'username does not exist');
        });
      });
    })['return'](username);
  });

  if (invalidateTokens) {
    promise = promise.tap(function addInvalidation(username) {
      return jwt.reset.call(this, username);
    });
  }

  if (remoteip) {
    promise = promise.tap(function resetLock(username) {
      return redis.del(redisKey(username, 'data', remoteip));
    });
  }

  return promise['return']({ success: true });
};