'use strict';

var Errors = require('common-errors');
var Promise = require('bluebird');
var jwt = Promise.promisifyAll(require('jsonwebtoken'));
var redisKey = require('./key.js');
var getMetadata = require('../actions/getMetadata.js');
var FlakeId = require('flake-idgen');
var flakeIdGen = new FlakeId();

/**
 * Logs user in and returns JWT and User Object
 * @param  {String}  username
 * @param  {String}  _audience
 * @return {Promise}
 */
exports.login = function login(username, _audience) {
  var redis = this.redis;
  var config = this.config;
  var jwtConfig = config.jwt;
  var algorithm = jwtConfig.hashingFunction;
  var defaultAudience = jwtConfig.defaultAudience;
  var secret = jwtConfig.secret;

  var audience = _audience || defaultAudience;

  // will have iat field, which is when this token was issued
  // we can check last access and verify the expiration date based on it
  var payload = {
    username: username,
    cs: flakeIdGen.next().toString('hex')
  };

  var token = jwt.sign(payload, secret, { algorithm: algorithm, audience: audience, issuer: 'ms-users' });

  if (audience !== defaultAudience) {
    audience = [audience, defaultAudience];
  } else {
    audience = [audience];
  }

  return Promise.props({
    lastAccessUpdated: redis.zadd(redisKey(username, 'tokens'), Date.now(), token),
    jwt: token,
    username: username,
    metadata: getMetadata.call(this, username, audience)
  }).then(function remap(props) {
    return {
      jwt: props.jwt,
      user: {
        username: props.username,
        metadata: props.metadata
      }
    };
  });
};

/**
 * Removes token if it is valid
 * @param  {String} token
 * @param  {String} audience
 * @return {Promise}
 */
exports.logout = function logout(token, audience) {
  var _this = this;

  var redis = this.redis;
  var config = this.config;
  var jwtConfig = config.jwt;
  var algorithm = jwtConfig.hashingFunction;
  var secret = jwtConfig.secret;
  var issuer = jwtConfig.issuer;

  return jwt.verifyAsync(token, secret, { issuer: issuer, audience: audience, algorithms: [algorithm] })['catch'](function (err) {
    _this.log.debug('error decoding token', err);
    throw new Errors.HttpStatusError(403, 'Invalid Token');
  }).then(function decodedToken(decoded) {
    return redis.zrem(redisKey(decoded.username, 'tokens'), token);
  })['return']({ success: true });
};

/**
 * Removes all issued tokens for a given user
 * @param {String} username
 */
exports.reset = function reset(username) {
  return this.redis.del(redisKey(username, 'tokens'));
};

/**
 * Verifies token and returns decoded version of it
 * @param  {String}       token
 * @param  {Array} audience
 * @param  {Boolean}      peek
 * @return {Promise}
 */
exports.verify = function verifyToken(token, audience, peek) {
  var _this2 = this;

  var redis = this.redis;
  var config = this.config;
  var jwtConfig = config.jwt;
  var algorithm = jwtConfig.hashingFunction;
  var secret = jwtConfig.secret;
  var ttl = jwtConfig.ttl;
  var issuer = jwtConfig.issuer;

  return jwt.verifyAsync(token, secret, { issuer: issuer, algorithms: [algorithm] })['catch'](function (err) {
    _this2.log.debug('invalid token passed: %s', token, err);
    throw new Errors.HttpStatusError(403, 'invalid token');
  }).then(function decodedToken(decoded) {
    if (audience.indexOf(decoded.aud) === -1) {
      throw new Errors.HttpStatusError(403, 'audience mismatch');
    }

    var username = decoded.username;

    var tokensHolder = redisKey(username, 'tokens');
    var lastAccess = redis.zscoreBuffer(tokensHolder, token).then(function getLastAccess(_score) {
      // parseResponse
      var score = parseInt(_score, 10);

      // throw if token not found or expired
      if (isNaN(score) || Date.now() > score + ttl) {
        throw new Errors.HttpStatusError(403, 'token has expired or was forged');
      }

      return score;
    });

    if (!peek) {
      lastAccess = lastAccess.then(function refreshLastAccess() {
        return redis.zadd(tokensHolder, Date.now(), token);
      });
    }

    return lastAccess['return'](decoded);
  });
};