'use strict';

var Errors = require('common-errors');
var Promise = require('bluebird'); // jshint ignore: line
var uuid = require('node-uuid');
var redisKey = require('../utils/key.js');
var crypto = require('crypto');
var URLSafeBase64 = require('urlsafe-base64');
var render = require('ms-mailer-templates');

/**
 * Encrypts buffer using alg and secret
 * @param  {String} algorithm
 * @param  {String} secret
 * @param  {Buffer} buffer
 * @return {Buffer}
 */
exports.encrypt = function encrypt(algorithm, secret, buffer) {
  var cipher = crypto.createCipher(algorithm, secret);
  var crypted = Buffer.concat([cipher.update(buffer), cipher.final()]);
  return crypted;
};

/**
 * Decrypts buffer using algoruthm and secret
 * @param  {String} algorithm
 * @param  {String} secret
 * @param  {Buffer} buffer
 * @return {Buffer}
 */
exports.decrypt = function decrypt(algorithm, secret, buffer) {
  var decipher = crypto.createDecipher(algorithm, secret);
  var dec = Buffer.concat([decipher.update(buffer), decipher.final()]);
  return dec;
};

/**
 * Generates complete link
 * @param  {Object} server
 * @param  {String} path
 * @return {String}
 */
exports.generateLink = function generateLink(server, path) {
  var proto = server.proto;
  var port = server.port;

  if (proto === 'http' && +port === 80 || proto === 'https' && +port === 443) {
    port = '';
  } else {
    port = ':' + port;
  }

  return proto + '://' + (server.host + port + path);
};

/**
 * Safely decodes
 * @param  {String} algorithm
 * @param  {String} secret
 * @param  {String} string
 * @return {Promise}
 */
exports.safeDecode = function safeDecode(algorithm, secret, string) {
  return Promise['try'](function decode() {
    return JSON.parse(exports.decrypt(algorithm, secret, URLSafeBase64.decode(string)));
  }).bind(this)['catch'](function remapError(err) {
    this.log.warn('cant decode token', err);
    throw new Errors.HttpStatusError(403, 'could not decode token');
  });
};

/**
 * Send an email with appropriate content
 *
 * @param  {String} email
 * @param  {String} type
 * @return {Promise}
 */
exports.send = function sendEmail(email) {
  var type = arguments.length <= 1 || arguments[1] === undefined ? 'activate' : arguments[1];
  var wait = arguments.length <= 2 || arguments[2] === undefined ? false : arguments[2];
  var redis = this.redis;
  var config = this.config;
  var mailer = this.mailer;
  var validation = config.validation;
  var server = config.server;
  var ttl = validation.ttl;
  var throttle = validation.throttle;
  var subjects = validation.subjects;
  var senders = validation.senders;
  var paths = validation.paths;
  var secret = validation.secret;
  var algorithm = validation.algorithm;
  var mailingAccount = validation.email;

  // method specific stuff
  var throttleEmailsKey = redisKey('vthrottle-' + type, email);
  var activationSecret = uuid.v4();
  var logger = this.log.child({ action: 'sendEmail', email: email });

  return redis.get(throttleEmailsKey).then(function isThrottled(reply) {
    if (reply) {
      throw new Errors.HttpStatusError(429, 'We\'ve already sent you an email, if it doesn\'t come - please try again in a little while or send us an email');
    }
  }).then(function generateContent() {
    // generate secret
    var enc = exports.encrypt(algorithm, secret, new Buffer(JSON.stringify({ email: email, token: activationSecret })));

    // generate context
    var context = {
      qs: '?q=' + URLSafeBase64.encode(enc)
    };

    switch (type) {
      case 'activate':
      case 'reset':
        context.link = exports.generateLink(server, paths[type]);
        break;
      default:
        throw new Errors.InvalidOperationError(type + ' action is not supported');
    }

    return render(type, context);
  }).then(function storeSecrets(emailTemplate) {
    var throttleArgs = [throttleEmailsKey, 1, 'NX'];
    if (throttle > 0) {
      throttleArgs.splice(2, 0, 'EX', throttle);
    }
    return redis.set(throttleArgs).then(function isThrottled(response) {
      if (!response) {
        throw new Errors.HttpStatusError(429, 'We\'ve already sent you an email, if it doesn\'t come - please try again in a little while or send us an email');
      }
    }).then(function updateSecret() {
      var secretKey = redisKey('vsecret-' + type, activationSecret);
      var args = [secretKey, email];
      if (ttl > 0) {
        args.push('EX', ttl);
      }
      return redis.set(args);
    })['return'](emailTemplate);
  }).then(function definedSubjectAndSend(emailTemplate) {
    var subject = undefined;
    var from = undefined;
    switch (type) {
      case 'activate':
      case 'reset':
        subject = subjects[type];
        from = senders[type];
        break;
      default:
        subject = '';
        from = 'noreply <support@example.com>';
    }

    var mail = {
      subject: subject,
      from: from,
      to: email,
      html: emailTemplate
    };

    var mailSent = mailer.send(mailingAccount, mail)['catch'](function mailingFailed(err) {
      logger.warn('couldn\'t send email', err);
    });

    if (wait) {
      return mailSent;
    }

    return { queued: true };
  });
};

/**
 * Verifies token
 * @param  {String}  string
 * @param  {String}  namespace
 * @param  {Boolean} expire
 * @return {Promise}
 */
exports.verify = function verifyToken(string, namespace, expires) {
  if (namespace === undefined) namespace = 'activate';
  var redis = this.redis;
  var config = this.config;
  var validation = config.validation;
  var validationSecret = validation.secret;
  var algorithm = validation.algorithm;

  return exports.safeDecode.call(this, algorithm, validationSecret, string).then(function inspectResult(message) {
    var email = message.email;
    var token = message.token;

    if (!email || !token) {
      throw new Errors.HttpStatusError(403, 'Decoded token misses references to email and/or secret');
    }

    var secretKey = redisKey('vsecret-' + namespace, token);
    return redis.get(secretKey).then(function inspectAssociatedData(associatedEmail) {
      if (!associatedEmail) {
        throw new Errors.HttpStatusError(404, 'token expired or is invalid');
      }

      if (associatedEmail !== email) {
        throw new Errors.HttpStatusError(412, 'associated email doesn\'t match token');
      }

      if (expires) {
        return redis.del(secretKey);
      }
    })['return'](email);
  });
};